//
//  YGGameScene.m
//  SpriteKit
//
//  Created by Lyle Lee on 2013. 12. 31..
//  Copyright (c) 2013년 Lyle Lee. All rights reserved.
//

#import "YGGameScene.h"

// 각 Node의 Physics Body를 판단하기 위한 카테고리 비트 마스크
typedef NS_ENUM(uint32_t, PhysicsBodyCategory) {
    PhysicsBodyCategorySpaceship = 0x01 << 0,
    PhysicsBodyCategoryRock = 0x01 << 1
};

@implementation YGGameScene

// 화면에 우주선을 그려준 후 액션을 추가함
-	(SKSpriteNode  *)newSpaceship
{
    // 우주선이 움직일 좌표들을 생성함
    CGPoint leftPoint = CGPointMake(30.0f, 50.0f);
    CGPoint rightPoint = CGPointMake(self.size.width - 30.0f, 50.0f);
    
    // 우주선을 표현할 Sprite Node 생성 및 초기 설정
    // spriteNodeWithImageNamed 메서드는 이미지 파일을 그대로 SKSpriteNode로 사용할 수 있도록 함
    SKSpriteNode *spaceShip = [SKSpriteNode spriteNodeWithImageNamed:@"Spaceship"];
    [spaceShip setSize:CGSizeMake(30.0f, 30.0f)];
    
    // 움직임 및 정지를 표현할 SKAction 객체를 생성함
    // 위에서 만들어준 rightPoint로 1.5로 동안 움직임
    SKAction *moveRight = [SKAction moveTo:rightPoint duration:1.5f];
    // 위에서 만들어준 leftPoint로 1.5로 동안 움직임
    SKAction *moveLeft = [SKAction moveTo:leftPoint duration:1.5f];
    // 1초 동안 잠시 움직임을 멈춤
    SKAction *wait = [SKAction waitForDuration:1.0f];
    
    // 연속된 Action을 표현하기 위해 배열로 Sequence를 생성함
    NSArray *sequenceArray = @[[wait copy], moveRight, [wait copy], moveLeft];
    
    // 연속된 새로운 Action을 생성함
    SKAction *hover = [SKAction sequence:sequenceArray];
    
    // 우주선 Sprite Node가 영원히 반복적으로 액션을 실행하도록 함
    [spaceShip runAction:[SKAction repeatActionForever:hover]];
    
    // 우주선에 물리 영역 설정
    [spaceShip setPhysicsBody:[SKPhysicsBody bodyWithRectangleOfSize:spaceShip.size]];
    
    // 우주선의 물리 영역 카테고리 설정
    [[spaceShip physicsBody] setCategoryBitMask:PhysicsBodyCategorySpaceship];
    
    // 우주선과 충돌 효과를 원하는 카테고리 설정
    [[spaceShip physicsBody] setContactTestBitMask:PhysicsBodyCategoryRock];
    
    // 우주선은 중력의 영향을 받으면 화면 아래로 떨어지므로, 중력의 영향을 받지 않도록 설정
    [[spaceShip physicsBody] setAffectedByGravity:NO];
    
    // 우주선의 질량이 운석과 같으면 충돌 효과가 너무 크므로 우주선에 운석이 튕겨나가도록 10배로 설정
    [[spaceShip physicsBody] setMass:10.0f];
    
    return spaceShip;
}

// 화면에 운석을 생성함
-	(void) addRock
{
    // 갈색 색상을 가진 8*8 사이즈의 네모 블록을 SKSpriteNode 객체로 생성함
    SKSpriteNode *rock = [[SKSpriteNode alloc] initWithColor:[SKColor brownColor] size:CGSizeMake(8.0f, 8.0f)];
    // 처음에 임의의 x 위치를 가지도록 설정해줌
    rock.position = CGPointMake(skRand(0, self.size.width), self.size.height - 50.0f);
    // rock 객체의 이름을 rock이라고 지정함. 이 이름은 SKScene에서 하위 node를 구분할 수 있는 역할을 해줌
    rock.name = @"rock";
    // 자신의 크기만큼 물리적 영역을 만들어줌
    rock.physicsBody = [SKPhysicsBody bodyWithRectangleOfSize:rock.size];
    
    // 운석의 물리영역 카테고리 설정
    [[rock physicsBody] setCategoryBitMask:PhysicsBodyCategoryRock];
    
    // 운석과 충돌을 원하는 카테고리 설정
    [[rock physicsBody] setContactTestBitMask:PhysicsBodyCategorySpaceship];
    
    // scene 위에 위치시킴
    [self addChild:rock];
}

// 임의의 CGFloat 값을 생성함
static inline CGFloat skRandf() {
    return rand() / (CGFloat) RAND_MAX;
}

// low와 high 사이의 임의의 CGFloat값을 생성함
static inline CGFloat skRand(CGFloat low, CGFloat high) {
    return skRandf() * (high - low) + low;
}

// Scene 안에 화면들을 전체적으로 그려줌
- (void)createSceneContents
{
    // 초기 화면 설정
    self.backgroundColor = [SKColor blackColor];
    self.scaleMode = SKSceneScaleModeAspectFit;
    
    // 화면에 우주선을 생성함
    SKSpriteNode *spaceship = [self newSpaceship];
    spaceship.position = CGPointMake(CGRectGetMidX(self.frame), CGRectGetMidY(self.frame) - 150);
    [self addChild:spaceship];
    
    SKAction *makeRocks = [SKAction sequence: @[[SKAction performSelector:@selector(addRock) onTarget:self], [SKAction waitForDuration:0.1f withRange:0.15f]]];
    
    [self runAction: [SKAction repeatActionForever:makeRocks]];
}

// UIViewController의 ViewDidAppear와 같은 역할을 함
- (void)didMoveToView:(SKView *)view
{
    [self createSceneContents];
    
    // 물체간 충돌시 제어하기 위한 델리게이트 설정
    [[self physicsWorld] setContactDelegate:self];
}

// 메모리 관리를 위하여 화면 영역 밖으로 벗어난 Node들은 메모리에서 제거함
- (void)didSimulatePhysics {
    [self enumerateChildNodesWithName:@"rock" usingBlock:^(SKNode *node, BOOL *stop) {
        if (node.position.y < 0.0f) {
            [node removeFromParent];
        }
    }];
}

#pragma mark -SKPysicsContact Delegate

// 물체의 충돌이 생길 때 호출되는 SKPhysicsContact 델리게이트 메서드
- (void)didBeginContact:(SKPhysicsContact *)contact
{
    // 카테고리 비트마스크로 충돌한 물체를 얻어냄
    SKSpriteNode *rockNode = nil;
    SKSpriteNode *spaceshipNode = nil;
    
    // SKPhysicsContact 객체에는 충돌된 두 물체 bodyA와 bodyB가 들어 있음
    // 두 물체 중 어떤 것이 운석이고 어떤 것이 우주선인지 if-else문을 통해 가려냄
    if (contact.bodyA.categoryBitMask == PhysicsBodyCategoryRock) {
        // bodyA가 운석일 때
        rockNode = (SKSpriteNode *)contact.bodyA.node;
        spaceshipNode = (SKSpriteNode *)contact.bodyB.node;
    }
    
    else
    {
        // bodyB가 운석일 때
        rockNode = (SKSpriteNode *)contact.bodyB.node;
        spaceshipNode = (SKSpriteNode *)contact.bodyA.node;
    }
    
    // 화면에서 운석 제거
    [rockNode removeFromParent];
    
    // 효과를 내줄 Emitter Load
    // 프로젝트 번들에 포함해준 Emitter 리소스 파일을 불러냄
    NSString *firePath = [[NSBundle mainBundle] pathForResource:@"Fire" ofType:@"sks"];
    SKEmitterNode *boomEmitter = [NSKeyedUnarchiver unarchiveObjectWithFile:firePath];
    
    // 우주선에 효과 올려줌
    [spaceshipNode addChild:boomEmitter];
    
    // Emitter에 사라지는 액션 추가
    [boomEmitter runAction:[SKAction fadeAlphaTo:1.0f duration:0.2f]];
    
    // 우주선에 0.2초 후 Emitter가 사라지게 액션 추가
    [spaceshipNode runAction:[SKAction waitForDuration:0.2f] completion:^{
        [boomEmitter removeFromParent];
    }];
}

@end
